/**************************************************************************\
|
|    Copyright (C) 2015 Marc Stevens
|    FOR MASTERMATH USE ONLY
|
\**************************************************************************/

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//#include "md5.h"

#define rotate_right(x,n) (((x)>>(n))|((x)<<(32-(n))))
#define rotate_left(x,n)  (((x)<<(n))|((x)>>(32-(n))))

uint32_t md5_ff(uint32_t b, uint32_t c, uint32_t d)
  { return d ^ (b & (c ^ d)); }
uint32_t md5_gg(uint32_t b, uint32_t c, uint32_t d)
  { return c ^ (d & (b ^ c)); }
uint32_t md5_hh(uint32_t b, uint32_t c, uint32_t d)
  { return b ^ c ^ d; }
uint32_t md5_ii(uint32_t b, uint32_t c, uint32_t d)
  { return c ^ (b | ~d); }
  
  
const unsigned md5_wt[] =
	{ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15
	, 1, 6, 11, 0, 5, 10, 15, 4, 9, 14, 3, 8, 13, 2, 7, 12
	, 5, 8, 11, 14, 1, 4, 7, 10, 13, 0, 3, 6, 9, 12, 15, 2
	, 0, 7, 14, 5, 12, 3, 10, 1, 8, 15, 6, 13, 4, 11, 2, 9 };

const uint32_t md5_iv[] = { 0x67452301, 0xefcdab89, 0x98badcfe, 0x10325476 };
const uint32_t md5_ac[] = 
	{ 0xd76aa478, 0xe8c7b756, 0x242070db, 0xc1bdceee, 0xf57c0faf, 0x4787c62a, 0xa8304613, 0xfd469501, 
	  0x698098d8, 0x8b44f7af, 0xffff5bb1, 0x895cd7be, 0x6b901122, 0xfd987193, 0xa679438e, 0x49b40821, 
	  0xf61e2562, 0xc040b340, 0x265e5a51, 0xe9b6c7aa, 0xd62f105d, 0x2441453 , 0xd8a1e681, 0xe7d3fbc8, 
	  0x21e1cde6, 0xc33707d6, 0xf4d50d87, 0x455a14ed, 0xa9e3e905, 0xfcefa3f8, 0x676f02d9, 0x8d2a4c8a, 
	  0xfffa3942, 0x8771f681, 0x6d9d6122, 0xfde5380c, 0xa4beea44, 0x4bdecfa9, 0xf6bb4b60, 0xbebfbc70, 
	  0x289b7ec6, 0xeaa127fa, 0xd4ef3085, 0x4881d05 , 0xd9d4d039, 0xe6db99e5, 0x1fa27cf8, 0xc4ac5665, 
	  0xf4292244, 0x432aff97, 0xab9423a7, 0xfc93a039, 0x655b59c3, 0x8f0ccc92, 0xffeff47d, 0x85845dd1, 
	  0x6fa87e4f, 0xfe2ce6e0, 0xa3014314, 0x4e0811a1, 0xf7537e82, 0xbd3af235, 0x2ad7d2bb, 0xeb86d391 };

const unsigned md5_rc[] = 
	{ 7, 12, 17, 22, 7, 12, 17, 22, 7, 12, 17, 22, 7, 12, 17, 22, 
	  5, 9 , 14, 20, 5, 9 , 14, 20, 5, 9 , 14, 20, 5, 9 , 14, 20, 
	  4, 11, 16, 23, 4, 11, 16, 23, 4, 11, 16, 23, 4, 11, 16, 23, 
	  6, 10, 15, 21, 6, 10, 15, 21, 6, 10, 15, 21, 6, 10, 15, 21 };

void md5compress(uint32_t ihv[4], const uint32_t block[16])
{
	uint32_t Q[68]; // everything is offset by 3 so Q_t is Q[3+t]
	uint32_t F,T,R;
	int i;
	Q[3-3] = ihv[0]; Q[3+0] = ihv[1]; Q[3-1] = ihv[2]; Q[3-2] = ihv[3]; ;
	for (i = 0; i < 16; ++i)
	{
		F=md5_ff(Q[3+i], Q[3+i-1], Q[3+i-2]);
		T=Q[3+i-3]+F+md5_ac[i]+block[md5_wt[i]];
		R=rotate_left(T,md5_rc[i]);
		Q[3+i+1] = Q[3+i] + R;
	}
	for (i = 16; i < 32; ++i)
	{
		F=md5_gg(Q[3+i], Q[3+i-1], Q[3+i-2]);
		T=Q[3+i-3]+F+md5_ac[i]+block[md5_wt[i]];
		R=rotate_left(T,md5_rc[i]);
		Q[3+i+1] = Q[3+i] + R;
	}
	for (i = 32; i < 48; ++i)
	{
		F=md5_hh(Q[3+i], Q[3+i-1], Q[3+i-2]);
		T=Q[3+i-3]+F+md5_ac[i]+block[md5_wt[i]];
		R=rotate_left(T,md5_rc[i]);
		Q[3+i+1] = Q[3+i] + R;
	}
	for (i = 48; i < 64; ++i)
	{
		F=md5_ii(Q[3+i], Q[3+i-1], Q[3+i-2]);
		T=Q[3+i-3]+F+md5_ac[i]+block[md5_wt[i]];
		R=rotate_left(T,md5_rc[i]);
		Q[3+i+1] = Q[3+i] + R;
	}
	ihv[0] += Q[3+61]; ihv[1] += Q[3+64]; ihv[2] += Q[3+63]; ihv[3] += Q[3+62];
}

void md5step(unsigned t, uint32_t Q[68], const uint32_t block[16])
{
	uint32_t F,T,R;
	if (t<16)      F=md5_ff(Q[3+t],Q[3+t-1],Q[3+t-2]);
	else if (t<32) F=md5_gg(Q[3+t],Q[3+t-1],Q[3+t-2]);
	else if (t<48) F=md5_hh(Q[3+t],Q[3+t-1],Q[3+t-2]);
	else           F=md5_ii(Q[3+t],Q[3+t-1],Q[3+t-2]);
	T=Q[3+t-3]+F+md5_ac[t]+block[md5_wt[t]];
	R=rotate_left(T,md5_rc[t]);
	Q[3+t+1] = Q[3+t] + R;
}

/* very small and fast random number generator, goes through 2^32*(2^32 - 1) states before output becomes repetitive */
uint32_t seed1, seed2;
uint32_t rng32()
{
	seed1 ^= seed1 << 13;
	seed1 ^= seed1 >> 17;
	return (seed1 ^= seed1 << 5) + (seed2 += 789456123);
}

int main(int argc, char** argv)
{
	/* declare all variables here */
	uint32_t ihv1[4], ihv2[4], ihv1out[4], ihv2out[4];
	uint32_t block[16];
	int i=0;
	
	/* seed random number generator */
	seed1=time(NULL);
	seed2=time(NULL);
	for (i=0; i < 100; ++i)
		rng32();

	/* try random chaining values (ihv) and message blocks (block) until a collision is found */
	while (1)
	{
		for (i=0; i < 4; ++i)
		{
			ihv1out[i]=ihv1[i]=rng32();
			ihv2out[i]=ihv2[i]=ihv1[i]+(1<<31);
		}	
		for (i=0; i < 16; ++i)
		{
			block[i] = rng32();
		}
		md5compress(ihv1out, block);
		md5compress(ihv2out, block);
		if (ihv1out[0] == ihv2out[0] && ihv1out[1]==ihv2out[1] && ihv1out[2]==ihv2out[2] && ihv1out[3]==ihv2out[3])
			break;
	}
	
	/* print the collision */
	printf("CV1in: ");
	for (i=0; i < 4; ++i)
		printf(" %08x", ihv1[i]);
	printf("\n");

	printf("CV2in: ");
	for (i=0; i < 4; ++i)
		printf(" %08x", ihv2[i]);
	printf("\n");

	printf("block: ");
	for (i=0; i < 16; ++i)
		printf(" %08x", block[i]);
	printf("\n");

	printf("CV1out: ");
	for (i=0; i < 4; ++i)
		printf(" %08x", ihv1out[i]);
	printf("\n");

	printf("CV2out: ");
	for (i=0; i < 4; ++i)
		printf(" %08x", ihv2out[i]);
	printf("\n");
}
